import time


class JuriggedDaemonSignaler:
    def __init__(self) -> None:
        self._do_cache_clear = False
        self._deleted_fixtures: set[str] = set()
        self._block_until: float | None = None

    def signal_clear_cache(self) -> None:
        self._do_cache_clear = True
        self._block_until = time.time() + 1

    def receive_clear_cache_signal(self) -> bool:
        ret = self._do_cache_clear
        cur_time = time.time()
        while self._block_until is not None and cur_time < self._block_until:
            time.sleep(self._block_until - cur_time)
            cur_time = time.time()
        self._block_until = None
        self._do_cache_clear = False
        return ret



"""
Pytest Hot Reloading plugin
"""
from __future__ import annotations

import inspect
import os
import sys
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Optional

from pytest_hot_reloading.client import PytestClient
from pytest_hot_reloading.jurigged_daemon_signalers import JuriggedDaemonSignaler

# this is modified by the daemon so that the pytest_collection hooks does not run
i_am_server = False

seen_paths: set[Path] = set()
signaler = JuriggedDaemonSignaler()

if TYPE_CHECKING:
    from pytest import Config, Item, Parser, Session


class EnvVariables(str, Enum):
    PYTEST_DAEMON_PORT = "PYTEST_DAEMON_PORT"
    PYTEST_DAEMON_PYTEST_NAME = "PYTEST_DAEMON_PYTEST_NAME"
    PYTEST_DAEMON_TIMEOUT = "PYTEST_DAEMON_TIMEOUT"
    PYTEST_DAEMON_WATCH_GLOBS = "PYTEST_DAEMON_WATCH_GLOBS"
    PYTEST_DAEMON_IGNORE_WATCH_GLOBS = "PYTEST_DAEMON_IGNORE_WATCH_GLOBS"
    PYTEST_DAEMON_START_IF_NEEDED = "PYTEST_DAEMON_START_IF_NEEDED"
    PYTEST_DAEMON_DISABLE = "PYTEST_DAEMON_DISABLE"
    PYTEST_DAEMON_DO_NOT_AUTOWATCH_FIXTURES = "PYTEST_DAEMON_DO_NOT_AUTOWATCH_FIXTURES"
    PYTEST_DAEMON_USE_WATCHMAN = "PYTEST_DAEMON_USE_WATCHMAN"


def pytest_addoption(parser) -> None:
    group = parser.getgroup("daemon")
    group.addoption(
        "--daemon",
        action="store_true",
        default=False,
        help="Start the daemon. If it is already running, the old instance will stop.",
    )
    group.addoption(
        "--daemon-disable",
        action="store_true",
        default=(os.getenv(EnvVariables.PYTEST_DAEMON_DISABLE, "False").lower() in ("true", "1")),
        help="Do not use the daemon for this test run.",
    )
    group.addoption(
        "--daemon-port",
        action="store",
        default=int(os.getenv(EnvVariables.PYTEST_DAEMON_PORT, "4852")),
        help="The port to use for the daemon. You generally shouldn't need to set this.",
    )
    group.addoption(
        "--pytest-name",
        action="store",
        default=os.getenv(EnvVariables.PYTEST_DAEMON_PYTEST_NAME, "pytest"),
        help="The name of the pytest executable or module. This is used for starting the daemon.",
    )
    group.addoption(
        "--daemon-timeout",
        action="store",
        default=os.getenv(EnvVariables.PYTEST_DAEMON_TIMEOUT, (5 * 60)),
        help="The timeout in seconds to wait on a test suite to finish. This is not yet implemented.",
    )
    group.addoption(
        "--daemon-watch-globs",
        action="store",
        default=os.getenv(EnvVariables.PYTEST_DAEMON_WATCH_GLOBS, "./*.py"),
        help="The globs to watch for changes. This is a colon separated list of globs.",
    )
    group.addoption(
        "--daemon-ignore-watch-globs",
        action="store",
        default=os.getenv(EnvVariables.PYTEST_DAEMON_IGNORE_WATCH_GLOBS, "./.venv/*"),
        help="The globs to ignore for changes. This is a colon separated list of globs.",
    )
    group.addoption(
        "--stop-daemon",
        action="store_true",
        default=False,
        help="Stop the daemon",
    )
    group.addoption(
        "--daemon-start-if-needed",
        action="store_true",
        default=(
            os.getenv(EnvVariables.PYTEST_DAEMON_START_IF_NEEDED, "False").lower()
            in ("true", "1")
        ),
        help=(
            "Start the daemon if it is not running. To use this with VS Code, "
            'you need add "python.experiments.optOutFrom": ["pythonTestAdapter"] to your config.'
        ),
    )
    group.addoption(
        "--daemon-do-not-autowatch-fixtures",
        action="store_true",
        default=(
            os.getenv(EnvVariables.PYTEST_DAEMON_DO_NOT_AUTOWATCH_FIXTURES, "False").lower()
            in ("true", "1")
        ),
        help=(
            "Do not automatically watch fixtures. "
            "Typically this would be used if there's too many fixtures and the watch glob is used instead."
        ),
    )
    group.addoption(
        "--daemon-use-watchman",
        action="store_true",
        default=(
            os.getenv(EnvVariables.PYTEST_DAEMON_USE_WATCHMAN, "False").lower() in ("true", "1")
        ),
        help=(
            "Use watchman instead of polling. "
            "This reduces CPU usage, takes up open file handles, and improves responsiveness. "
            "Some systems cannot reliably use this."
        ),
    )


# list of pytest hooks
# https://docs.pytest.org/en/stable/reference.html#_pytest.hookspec.pytest_addhooks


def do_early_escape(config: Config) -> bool:
    if getattr(config.option, "collectonly", None):
        return True
    if i_am_server:
        return True
    if getattr(config.option, "help", None):
        return True
    if config.option.daemon_disable:
        return True
    return False


def pytest_load_initial_conftests(early_config: Config, args: list[str], parser: Parser) -> None:
    if do_early_escape(early_config):
        return
    early_config.known_args_namespace.noconftest = True


def pytest_cmdline_main(config: Config) -> Optional[int]:
    """
    This hook is called by pytest and is one of the first hooks.
    """
    if do_early_escape(config):
        return None
    status_code = _plugin_logic(config)
    # dont do any more work. Don't let pytest continue
    return status_code  # status code 0


fixture_names: set[str] = set()


def monkey_patch_jurigged_function_definition():
    import jurigged.codetools as jurigged_codetools  # type: ignore
    import jurigged.utils as jurigged_utils  # type: ignore

    OrigFunctionDefinition = jurigged_codetools.FunctionDefinition
    OrigDeleteOperation = jurigged_codetools.DeleteOperation

    import ast

    class NewDeleteOperation(OrigDeleteOperation):
        def __init__(self, *args, **kwargs) -> None:
            super().__init__(*args, **kwargs)

            self._signal_clear_cache_if_fixture()

        def _signal_clear_cache_if_fixture(self) -> None:
            """
            Clear the cache if a fixture is deleted.

            If this isn't here, then deleted fixtures may still exist.
            """
            if self.defn.name in fixture_names:
                signaler.signal_clear_cache()

    class NewFunctionDefinition(OrigFunctionDefinition):
        def reevaluate(self, new_node, glb):
            is_test = new_node.name.startswith("test_")
            if is_test:
                if not hasattr(self.node, "args"):
                    is_test = False
                else:
                    old_sig = [x.arg for x in self.node.args.args]
            else:
                if new_node.name in fixture_names:
                    # if a fixture is updated, then clear the session cache to avoid stale responses
                    signaler.signal_clear_cache()
            # monkeypatch: The assertion rewrite is from pytest. Jurigged doesn't
            #              seem to have a way to add rewrite hooks
            new_node = self.apply_assertion_rewrite(new_node, glb)
            obj = super().reevaluate(new_node, glb)

            if is_test:
                new_sig = [x.arg for x in new_node.args.args]

            # if the signature changes, clear the session cache
            # otherwise pytest will use the old signature.
            # This is a more of a band-aid because the session cache
            # could be more intelligently updated based on what was changed.
            # This band-aid fixes tests using stale fixture info, which
            # can result in unpredictable behavior that requires
            # restarting the daemon.
            if is_test:
                if old_sig != new_sig:
                    signaler.signal_clear_cache()

            return obj

        def apply_assertion_rewrite(self, ast_func, glb):
            from _pytest.assertion.rewrite import AssertionRewriter

            nodes: list[ast.AST] = [ast_func]  # type: ignore
            while nodes:
                node = nodes.pop()
                for name, field in ast.iter_fields(node):
                    if isinstance(field, list):
                        new: list[ast.AST] = []  # type: ignore
                        for i, child in enumerate(field):
                            if isinstance(child, ast.Assert):
                                # Transform assert.
                                new.extend(
                                    AssertionRewriter(glb["__file__"], None, None).visit(child)
                                )
                            else:
                                new.append(child)
                                if isinstance(child, ast.AST):
                                    nodes.append(child)
                        setattr(node, name, new)
                    elif (
                        isinstance(field, ast.AST)
                        # Don't recurse into expressions as they can't contain
                        # asserts.
                        and not isinstance(field, ast.expr)
                    ):
                        nodes.append(field)
            return ast_func

        def stash(self, lineno=1, col_offset=0):
            # monkeypatch: There's an off-by-one bug coming from somewhere in jurigged.
            #              This affects replaced functions. When line numbers are wrong
            #              the debugger and inspection logic doesn't work as expected.
            if not isinstance(self.parent, OrigFunctionDefinition):
                co = self.get_object()
                if co and (delta := lineno - co.co_firstlineno):
                    delta -= 1  # fix off-by-one
                    if delta != 0:
                        self.recode(jurigged_utils.shift_lineno(co, delta), use_cache=False)

            return super(OrigFunctionDefinition, self).stash(lineno, col_offset)

    # monkey patch in new definition
    jurigged_codetools.FunctionDefinition = NewFunctionDefinition
    jurigged_codetools.DeleteOperation = NewDeleteOperation


def monkeypatch_group_definition():
    import jurigged.codetools as jurigged_codetools  # type: ignore

    def append(self, *children, ensure_separation=False):
        for child in children:
            # ensure_separation creates line number diff
            # an example where this was a problem:
            #
            # 15 class MyClass:
            # 77     do_something()  # type: ignore  <--- blank line inserted between do_something() and comment
            # 78
            # 79     def my_func(...)  <--- becomes line 80
            #
            # the monkey patch removes it
            #
            # removed code:
            # if (
            #     ensure_separation
            #     and self.children
            #     and not self.children[-1].well_separated(child)
            # ):
            #     ws = LineDefinition(
            #         node=None, text="\n", filename=self.filename
            #     )
            #     self.children.append(ws)
            #     ws.set_parent(self)
            self.children.append(child)
            child.set_parent(self)

    jurigged_codetools.GroupDefinition.append = append


def setup_jurigged(config: Config):
    import jurigged

    monkey_patch_jurigged_function_definition()
    monkeypatch_group_definition()
    if not config.option.daemon_do_not_autowatch_fixtures:
        monkeypatch_fixture_marker(config.option.daemon_use_watchman)
    else:
        print("Not autowatching fixtures")

    pattern = _get_pattern_filters(config)
    # TODO: intelligently use poll versus watchman (https://github.com/JamesHutchison/pytest-hot-reloading/issues/16)
    jurigged.watch(
        pattern=pattern,
        poll=(not config.option.daemon_use_watchman),
    )


def watch_file(path: Path | str) -> None:
    from jurigged import registry
    from jurigged.utils import glob_filter

    registry.auto_register(filter=glob_filter(str(path)))


seen_files: set[str] = set()


def monkeypatch_fixture_marker(use_watchman: bool):
    import pytest
    from _pytest import fixtures

    FixtureFunctionMarkerOrig = fixtures.FixtureFunctionMarker

    # FixtureFunctionMarker is marked as final
    class FixtureFunctionMarkerNew(FixtureFunctionMarkerOrig):  # type: ignore # noqa
        def __call__(self, func, *args, **kwargs):
            fixture_names.add(func.__name__)

            return super().__call__(func, *args, **kwargs)

    fixture_original = pytest.fixture

    # doing a pure class only monkeypatch was breaking event_loop fixture
    # so patching out fixture function to use new class
    def _new_fixture(*args, **kwargs):
        # get current file where this is called
        frame = inspect.stack()[1]
        module = inspect.getmodule(frame[0])
        fixture_file = module.__file__

        # add fixture file to watched files
        if fixture_file not in seen_files:
            seen_files.add(fixture_file)
            watch_file(fixture_file)

        fixtures.FixtureFunctionMarker = FixtureFunctionMarkerNew
        try:
            ret = fixture_original(*args, **kwargs)
        finally:
            fixtures.FixtureFunctionMarker = FixtureFunctionMarkerOrig

        return ret

    pytest.fixture = _new_fixture


def _plugin_logic(config: Config) -> int:
    """
    The core plugin logic. This is where it splits based on whether we are the server or client.

    In either case, the pytest logic will not continue after this.
    """
    # if daemon is passed, then we are the daemon / server
    # if daemon is not passed, then we are the client
    daemon_port = int(config.option.daemon_port)  # --daemon-port
    if config.option.daemon:  # --daemon
        # pytest prints out "collecting ...". The leading \r prevents that
        print("\rStarting daemon...")
        setup_jurigged(config)

        from pytest_hot_reloading.daemon import PytestDaemon

        daemon = PytestDaemon(daemon_port=daemon_port, signaler=signaler)

        daemon.run_forever()
        sys.exit(0)
    else:
        pytest_name = config.option.pytest_name  # --pytest-name
        client = PytestClient(
            daemon_port=daemon_port,
            pytest_name=pytest_name,
            start_daemon_if_needed=config.option.daemon_start_if_needed,  # --daemon-start-if-needed
            do_not_autowatch_fixtures=config.option.daemon_do_not_autowatch_fixtures,  # --daemon-do-not-autowatch-fixtures
            use_watchman=config.option.daemon_use_watchman,  # --daemon-use-watchman
            additional_args=config.invocation_params.args,
        )

        if config.option.stop_daemon:  # --stop-daemon
            client.stop()
            return 0

        cwd = config.invocation_params.dir
        args = list(config.invocation_params.args)

        status_code = client.run(cwd, args)
        return status_code


def _get_pattern_filters(config: Config) -> str | Callable[[str], bool]:
    """
    Jurigged takes in a pattern argument. The argument is either a glob string
    or a function that returns True if the path passed into it should be watched.

    This creates a function filter that will return True if the path matches.

    The logic takes in the --daemon-watch-globs and --daemon-ignore-watch-globs options
    and creates a function that will return True if the path matches the watch globs.

    The seen_paths set is used to prevent duplicate paths from being watched and also
    acts as a short circuit for paths that have already been seen.
    """
    global seen_paths

    import fnmatch
    import re

    def normalize(glob: str) -> str:
        if glob.startswith("~"):
            glob = os.path.expanduser(glob)
        elif not glob.startswith("/"):
            glob = os.path.abspath(glob)
        if os.path.isdir(glob):
            glob = os.path.join(glob, "*")
        return glob

    watch_globs = config.option.daemon_watch_globs.split(":")
    regex_matches = [re.compile(fnmatch.translate(normalize(glob))).match for glob in watch_globs]

    ignore_watch_globs = config.option.daemon_ignore_watch_globs
    if ignore_watch_globs:
        ignore_globs = config.option.daemon_ignore_watch_globs.split(":")
        ignore_regex_matches = [
            re.compile(fnmatch.translate(normalize(glob))).match for glob in ignore_globs
        ]
    else:
        ignore_regex_matches = []

    def matcher(filename: str) -> bool:
        if filename in seen_paths:
            return False
        seen_paths.add(Path(filename))
        if any(regex_match(filename) for regex_match in regex_matches):
            if not any(
                ignore_regex_match(filename) for ignore_regex_match in ignore_regex_matches
            ):
                return True
        return False

    return matcher


def pytest_collection_modifyitems(session: Session, config: Config, items: list[Item]) -> None:
    """
    This hook is called by pytest after the collection phase.

    This adds tests and conftests to the watch list automatically.

    The client should never get this far. This should only be
    used by the daemon.
    """
    global seen_paths

    for item in items:
        if item.path and item.path not in seen_paths:
            watch_file(item.path)
            seen_paths.add(item.path)



import copy
import json
import os
import re
import socket
import subprocess
import sys
import tempfile
import time
import traceback
from pathlib import Path
from threading import Thread
from typing import Counter, Generator, Sequence
from xmlrpc.server import SimpleXMLRPCServer

import pytest
from cachetools import TTLCache

from pytest_hot_reloading.jurigged_daemon_signalers import JuriggedDaemonSignaler
from pytest_hot_reloading.workarounds import (
    run_workarounds_post,
    run_workarounds_pre,
)


class PytestDaemon:
    def __init__(
        self,
        signaler: JuriggedDaemonSignaler,
        daemon_host: str = "localhost",
        daemon_port: int = 4852,
    ) -> None:
        self._daemon_host = daemon_host
        self._daemon_port = daemon_port
        self._server: SimpleXMLRPCServer | None = None
        self._signaler = signaler

    @property
    def pid_file(self) -> Path:
        return Path(tempfile.gettempdir()) / f".pytest_hot_reloading_{self._daemon_port}.pid"

    @staticmethod
    def start(
        host: str,
        port: int,
        pytest_name: str = "pytest",
        watch_globs: str | None = None,
        ignore_watch_globs: str | None = None,
        do_not_autowatch_fixtures: bool | None = None,
        use_watchman: bool | None = None,
        additional_args: Sequence[str] | None = None,
    ) -> None:
        # start the daemon such that it will not close when the parent process closes
        if host == "localhost":
            args = [
                sys.executable,
                "-m",
                pytest_name,
                "--daemon",
                "--daemon-port",
                str(port),
            ]
            if watch_globs:
                args += ["--daemon-watch-globs", watch_globs]
            if ignore_watch_globs:
                args += ["--daemon-ignore-watch-globs", ignore_watch_globs]
            if do_not_autowatch_fixtures:
                args += ["--daemon-do-not-autowatch-fixtures"]
            if use_watchman:
                args += ["--daemon-use-watchman"]
            subprocess.Popen(
                args + list(additional_args or []),
                env=os.environ,
                cwd=os.getcwd(),
            )
        else:
            raise NotImplementedError("Only localhost is supported for now")
        PytestDaemon.wait_to_be_ready(host, port)

    def stop(self) -> dict:
        if self._server:
            t = Thread(target=self._server.shutdown, daemon=True)
            t.start()
            self._delete_pid_file()

        return {"shutdown": "ok"}

    @staticmethod
    def wait_to_be_ready(host: str = "localhost", port: int = 4852) -> None:
        # poll the connection to the daemon using sockets
        # and return when it is ready
        for _ in range(100):
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.connect((host, port))
            except ConnectionRefusedError:
                time.sleep(0.1)
                continue
            else:
                break
        else:
            raise Exception("Could not connect to the daemon")

    def run_forever(self) -> None:  # create an XML-RPC server
        try:
            server = SimpleXMLRPCServer((self._daemon_host, self._daemon_port))
        except OSError as err:
            if "Address already in use" in str(err):
                self._kill_existing_daemon()
                time.sleep(2)
                server = SimpleXMLRPCServer((self._daemon_host, self._daemon_port))

        self._write_pid_file()

        # register the 'run_pytest' function
        server.register_function(self.run_pytest, "run_pytest")  # type: ignore
        server.register_function(self.stop, "stop")

        self._server = server
        server.serve_forever()

    def _write_pid_file(self) -> None:
        with open(self.pid_file, "w") as f:
            f.write(str(os.getpid()))

    def _delete_pid_file(self) -> None:
        if os.path.exists(self.pid_file):
            os.unlink(self.pid_file)

    def _kill_existing_daemon(self) -> None:
        try:
            with open(self.pid_file, "r") as f:
                pid = int(f.read())
            os.kill(pid, 9)
        except FileNotFoundError:
            raise Exception(f"Port {self._daemon_port} is already in use")

    def run_pytest(self, cwd: str, env_json: str, sys_path: list[str], args: list[str]) -> dict:
        try:
            # run pytest using command line args
            # run the pytest main logic
            in_progress_workarounds = self._workaround_library_issues_pre()

            import pytest_hot_reloading.plugin as plugin

            # indicate to the plugin to NOT run custom pytest collect logic
            plugin.i_am_server = True

            # capture stdout and stderr
            # and return the output
            import io

            stdout = io.StringIO()
            stderr = io.StringIO()

            # backup originals
            stdout_bak = sys.stdout
            stderr_bak = sys.stderr

            sys.stdout = stdout
            sys.stderr = stderr

            if self._signaler.receive_clear_cache_signal():
                session_item_cache.clear()

            import _pytest.main

            # monkeypatch in the main that does test collection caching
            orig_main = _pytest.main._main
            _pytest.main._main = _pytest_main

            # switch to client working directory
            # do NOT store and restore previous because it might disappear and create errors
            os.chdir(cwd)

            # copy the environment
            env_old = os.environ.copy()
            # switch to client environment
            new_env = json.loads(env_json)
            os.environ.update(new_env)

            # copy sys.path
            sys_path_old = sys.path
            # switch to client path
            sys.path = sys_path

            try:
                # args must omit the calling program
                status_code = pytest.main(["--color=yes"] + args)
            finally:
                self._workaround_library_issues_post(in_progress_workarounds)

                # restore sys.path
                sys.path = sys_path_old

                # restore environment
                os.environ.update(env_old)

                # restore originals
                _pytest.main._main = orig_main

                sys.stdout = stdout_bak
                sys.stderr = stderr_bak

                stdout.seek(0)
                stderr.seek(0)
                stdout_str = stdout.read()
                stderr_str = stderr.read()

                print(stdout_str, file=sys.stdout)
                print(stderr_str, file=sys.stderr)
            return {
                "stdout": self._remove_ansi_escape(stdout_str).encode("utf-8"),
                "stderr": self._remove_ansi_escape(stderr_str).encode("utf-8"),
                "status_code": int(status_code),
            }
        except Exception:
            return {
                "stdout": b"",
                "stderr": traceback.format_exc().encode("utf-8"),
                "status_code": -1,
            }

    def _remove_ansi_escape(self, s: str) -> str:
        return re.sub(r"\x1b(\[.*?[@-~]|\].*?(\x07|\x1b\\))", "", s, flags=re.MULTILINE)

    def _workaround_library_issues_pre(self) -> list[Generator]:
        return run_workarounds_pre()

    def _workaround_library_issues_post(self, in_progress_workarounds: list[Generator]) -> None:
        run_workarounds_post(in_progress_workarounds)


session_item_cache: TTLCache[tuple, tuple] = TTLCache(16, 500)
# hack: keeping a session cache since pytest has session references
#       littered everywhere on objects
prior_sessions: set[pytest.Session] = set()


def _manage_prior_session_garbage(session: pytest.Session) -> None:
    """
    Pytest creates a bunch of objects and nodes and assigns the session
    to them. This creates a lot of dangling references to sessions that
    can come up later due to reuse. To work around this, all prior
    sessions have their dicts updated to point to the latest session.

    To avoid accumulating too many sessions and taking up memory as well
    as runtime, this cleans out all the sessions that appear to be redundant.

    This isn't quite perfect but should maybe be enough. There's likely some
    corner cases where the session with the fewest references doesn't meet the
    count requirement. There may also be cases where the smallest session
    has something important on it and we don't want to clean it up, but it
    gets cleaned up anyways.

    The use case that drew attention to this problem was an autouse session
    fixture. The fixture's request object was referencing the session that
    used the fixture, which at some point in the flow creates a problem
    because that old session is not properly set up anymore.
    """
    ref_counts = {
        prior_session: sys.getrefcount(prior_session) for prior_session in prior_sessions
    }
    counts = Counter(ref_counts.values())
    min_count = min(counts.keys()) if ref_counts else 0

    for prior_session in list(prior_sessions):
        if (
            len(prior_sessions) > 5
            and counts[min_count] > 3
            and ref_counts[prior_session] <= min_count
        ):
            prior_sessions.remove(prior_session)
        else:
            prior_session.__dict__ = session.__dict__


# performance improvements
# when doing a best_effort_copy, do not copy these attributes
no_copy = {
    "_arg2fixturedefs",
    "_fixtureinfo",
    "keywords",
    "_fixturemanager",
    "_pyfuncitem",
}

# when doing a best_effort_copy, do not deep copy this
# instead, force a best effort up to a given depth
use_best_effort_copy = {
    "_request",
}


def _pytest_main(config: pytest.Config, session: pytest.Session):
    """
    A monkey patched version of _pytest._main that caches test collection
    """
    _manage_prior_session_garbage(session)

    import _pytest.capture

    _pytest.capture.CaptureManager.stop_global_capturing = lambda self: None  # type: ignore
    start_global_capturing = _pytest.capture.CaptureManager.start_global_capturing
    resume_global_capture = _pytest.capture.CaptureManager.resume_global_capture

    def start_global_capture_if_needed(self: _pytest.capture.CaptureManager):
        if self._global_capturing is None:
            start_global_capturing(self)
        return resume_global_capture(self)

    _pytest.capture.CaptureManager.resume_global_capture = start_global_capture_if_needed  # type: ignore

    def best_effort_copy(item, depth_remaining=2, force_best_effort=False):
        """
        Copy test items. The items have references to modules and
        other things that cannot be deep copied.
        """
        if depth_remaining <= 0:
            return item
        try:
            item_copy = copy.copy(item)
        except TypeError:
            return item
        # NodeKeywords is an example of an object without a __dict__
        if hasattr(item, "__dict__"):
            for k, v in item.__dict__.items():
                # performance-tweaks
                if k in no_copy:
                    item_copy.__dict__[k] = v
                    continue
                if k in use_best_effort_copy:
                    item_copy.__dict__[k] = best_effort_copy(v, 2, force_best_effort=True)
                    continue
                if force_best_effort:
                    item_copy.__dict__[k] = best_effort_copy(
                        v, depth_remaining - 1, force_best_effort=True
                    )
                    continue

                try:
                    item_copy.__dict__[k] = copy.deepcopy(v)
                except KeyboardInterrupt:
                    raise
                except TypeError:
                    # Non-pickelable objects
                    item_copy.__dict__[k] = best_effort_copy(v, depth_remaining - 1)
        return item_copy

    num_tests_collected: int

    # here config.args becomes basically the tests to run. Other arguments are omitted
    # not 100% sure this is always the case
    session_key = tuple(config.args)
    try:
        items = session_item_cache[session_key]
    except KeyError:
        # not in the cache, do test collection
        start = time.time()
        config.hook.pytest_collection(session=session)
        print(f"Pytest Daemon: Collection took {(time.time() - start):0.3f} seconds")
        session_item_cache[session_key] = tuple(best_effort_copy(x) for x in session.items)
        num_tests_collected = session.testscollected
    else:
        print("Pytest Daemon: Using cached collection")
        # Assign the prior test items (tests to run) and config to the current session
        session.items = items  # type: ignore
        num_tests_collected = len(items)
        session.config = config
        for i in items:
            # Items have references to the config and the session
            i.config = config
            i.session = session
            if i._request:
                i._request._pyfuncitem = i
    config.hook.pytest_runtestloop(session=session)
    prior_sessions.add(session)

    if session.testsfailed:
        return pytest.ExitCode.TESTS_FAILED
    elif num_tests_collected == 0:
        return pytest.ExitCode.NO_TESTS_COLLECTED
    return None



from typing import Callable, Generator, NamedTuple, Optional

Workaround = NamedTuple(
    "Workaround", [("module", str), ("func", Callable[[], Optional[Generator]])]
)

workarounds: list[Workaround] = []


def register_workaround(module_name: str):
    def _register_workaround(func: Callable[[], Optional[Generator]]) -> None:
        """
        Register a workaround. A workaround is a function that takes in
        a list of the arguments passed into pytest. The function may
        be a generator function, and if so, yield will separate the pre
        and post calls.
        """
        workarounds.append(Workaround(module_name, func))

    return _register_workaround


@register_workaround("xdist")
def xdist_workaround() -> None:
    """
    pytest-xdist is not supported. The test collection behaves differently
    and some libraries such as pytest-django may have bugs when its enabled.
    """
    from xdist import plugin  # type: ignore

    # monkey patch to force zero processes
    plugin.parse_numprocesses = lambda s: None


@register_workaround("pytest_django")
def pytest_django_tox_workaround() -> None:
    """
    pytest-django will attempt to add a suffix and they will accumulate with each run.
    Note that running tox with hot reloading doesn't make sense anyways.
    """
    from pytest_django import fixtures

    # monkey patch to disable suffix logic used by xdist and tox
    fixtures._set_suffix_to_test_databases = lambda suffix: None


@register_workaround("vscode_pytest")
def vscode_pytest_workaround() -> None:
    """
    VS Code seems to keep track of collected tests and does a check to ensure
    they were not already collected before returning them.

    Also, it uses a list when all it does is an existence check, so this swaps
    the list for a set with `append`.
    """

    import vscode_pytest  # type: ignore

    class SetWithAppendRemove(set):
        def append(self, val) -> None:
            self.add(val)

        def remove(self, val) -> None:
            self.remove(val)

    vscode_pytest.collected_tests_so_far = SetWithAppendRemove()


def run_workarounds_pre() -> list[Generator]:
    in_progress_workarounds = []
    for module_name, workaround in workarounds:
        try:
            __import__(module_name)
        except ImportError:
            continue  # not installed
        try:
            __import__(f"{module_name}._clear_hot_reload_workarounds")
        except ImportError:
            pass
        else:
            continue  # workaround no longer needed
        result = workaround()
        if result is not None:
            next(result)
            in_progress_workarounds.append(result)
    return in_progress_workarounds


def run_workarounds_post(in_progress_workarounds: list[Generator]) -> None:
    for workaround in in_progress_workarounds:
        try:
            next(workaround)
        except StopIteration:
            pass



import json
import os
import socket
import sys
import time
import xmlrpc.client
from pathlib import Path
from typing import Sequence, cast


class PytestClient:
    _socket: socket.socket | None
    _daemon_host: str
    _daemon_port: int
    _pytest_name: str
    _will_start_daemon_if_needed: bool
    _do_not_autowatch_fixtures: bool
    _use_watchman: bool

    def __init__(
        self,
        daemon_host: str = "localhost",
        daemon_port: int = 4852,
        pytest_name: str = "pytest",
        start_daemon_if_needed: bool = False,
        do_not_autowatch_fixtures: bool = False,
        use_watchman: bool = False,
        additional_args: Sequence[str] = [],
    ) -> None:
        self._socket = None
        self._daemon_host = daemon_host
        self._daemon_port = daemon_port
        self._pytest_name = pytest_name
        self._will_start_daemon_if_needed = start_daemon_if_needed
        self._do_not_autowatch_fixtures = do_not_autowatch_fixtures
        self._use_watchman = use_watchman
        self._additional_args = additional_args

    def _get_server(self) -> xmlrpc.client.ServerProxy:
        server_url = f"http://{self._daemon_host}:{self._daemon_port}"
        server = xmlrpc.client.ServerProxy(server_url)

        return server

    def run(self, cwd: Path, args: list[str]) -> int:
        if self._will_start_daemon_if_needed:
            self._start_daemon_if_needed()
        elif not self._daemon_running():
            raise Exception(
                "Daemon is not running and must be started, or add --daemon-start-if-needed"
            )

        server = self._get_server()

        env = os.environ.copy()
        sys_path = sys.path

        start = time.time()
        result: dict = cast(dict, server.run_pytest(str(cwd), json.dumps(env), sys_path, args))
        print(f"Daemon took {(time.time() - start):.3f} seconds to reply")

        stdout = result["stdout"].data.decode("utf-8")
        stderr = result["stderr"].data.decode("utf-8")

        print(stdout, file=sys.stdout)
        print(stderr, file=sys.stderr)

        return result["status_code"]

    def stop(self) -> None:
        """
        Stop the daemon
        """
        server = self._get_server()

        try:
            server.stop()
        except OSError:
            print("Daemon is not running")
        else:
            print("Daemon stopped")

    def abort(self) -> None:
        # Close the socket
        if self._socket:
            self._socket.close()

    def _daemon_running(self) -> bool:
        # first, try to connect
        try:
            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._socket.connect((self._daemon_host, self._daemon_port))
            # the daemon is running
            # close the socket
            self._socket.close()
            return True
        except ConnectionRefusedError:
            # the daemon is not running
            return False

    def _start_daemon_if_needed(self) -> None:
        # check if the daemon is running on the expected host and port
        # if not, start the daemon
        if not self._daemon_running():
            self._start_daemon()

    def _start_daemon(self) -> None:
        from pytest_hot_reloading.daemon import PytestDaemon

        # start the daemon
        PytestDaemon.start(
            host=self._daemon_host,
            port=self._daemon_port,
            pytest_name=self._pytest_name,
            do_not_autowatch_fixtures=self._do_not_autowatch_fixtures,
            use_watchman=self._use_watchman,
            additional_args=self._additional_args,
        )
